import asyncio

import pytest
from conftest import vcr_c

from aiotx.clients import AioTxBTCClient


@pytest.mark.mysql
@vcr_c.use_cassette("btc/test_async_monitoring_mysql.yaml")
async def test_async_monitoring_mysql(btc_client_mysql: AioTxBTCClient):
    blocks = []
    transactions = []

    @btc_client_mysql.monitor.on_block
    async def handle_block(block):
        blocks.append(block)

    @btc_client_mysql.monitor.on_transaction
    async def handle_transaction(transaction):
        transactions.append(transaction)

    await btc_client_mysql.import_address("tb1pmuuv2qjujv9qawqcc424nhlnmux7mvsyqj7qgc6z0vwqujvx4k9s34kuxn", 2811502)
    await btc_client_mysql.import_address("tb1putn7zkfjr97xd77as09w5syy3x7xr5crp99wycjcxgdwrq3le9lse9fdkd")
    await btc_client_mysql.import_address("tb1p2z70my9zwxsuqevdxjn909c7jyp89z4qv3uhy5wggn47ruvfa65s4fnngd")
    await btc_client_mysql.import_address("tb1paf6damf5052arl3r2lsufuhyu48yth8mrgxdtqj9pervz624q6xqxm7ew3")
    await btc_client_mysql.monitor._add_new_utxo(
        "tb1paf6damf5052arl3r2lsufuhyu48yth8mrgxdtqj9pervz624q6xqxm7ew3",
        "8b88ebe1d0bb1eda7b1512853e75d3884b99e6634317589b1f7457ab92f1e41f",
        499824213,
        0,
    )
    start_balance_utxo_wallet = await btc_client_mysql.get_balance(
        "tb1paf6damf5052arl3r2lsufuhyu48yth8mrgxdtqj9pervz624q6xqxm7ew3"
    )
    assert start_balance_utxo_wallet == 499824213
    await btc_client_mysql.start_monitoring()

    try:
        await asyncio.sleep(4)
    except KeyboardInterrupt:
        btc_client_mysql.stop_monitoring()

    assert len(blocks) > 0
    assert len(transactions) > 0

    assert 2811502 in blocks
    assert 2811503 in blocks
    assert 2811504 in blocks

    # Checking we have all tb1putn7zkfjr97xd77as09w5syy3x7xr5crp99wycjcxgdwrq3le9lse9fdkd transaction
    assert "3f419799747dc83a33a84a8d58fe5e9265d9d1b9b122de6914eb849abb04d9ae" in [tx["txid"] for tx in transactions]
    assert "45fabc1d7bb2b1255d78a00e1daebe37ed855b7965fbae6732bf0624e426d30d" in [tx["txid"] for tx in transactions]
    assert "6643bcac7025154b1ad35d92da73acf7dc51df5d80fab23d9734b07600d8b615" in [tx["txid"] for tx in transactions]

    # tb1p2z70my9zwxsuqevdxjn909c7jyp89z4qv3uhy5wggn47ruvfa65s4fnngd transactions
    assert "c5ff30d25de95338bc4436c6fd49ce35bf74297671f9d0a5a8d52b5022623cc7" in [tx["txid"] for tx in transactions]
    assert "1eb37d5cbad786b6c2954b5d6f69df62a7294a13d068a1393ce4869f61c8b293" in [tx["txid"] for tx in transactions]

    # tb1pmuuv2qjujv9qawqcc424nhlnmux7mvsyqj7qgc6z0vwqujvx4k9s34kuxn transaction
    assert "56f677b37c44f9609f6d2ee121effe5a59b0c550c9be5073d03d8c365dde0e83" in [tx["txid"] for tx in transactions]
    assert "0cfe0b554f766ed374a426497a9163a3beb9732e799ffa623894efaea81a4415" in [tx["txid"] for tx in transactions]
    assert "5f8c9b318ab9156598d3d04701c08f374832bb1ce8704e629a68961b08d290c1" in [tx["txid"] for tx in transactions]
    assert "a9bfc94a0a5d7e7e46f1626b9d1848da96cc85689528ca3187e55164be8f3d04" in [tx["txid"] for tx in transactions]


    balance = await btc_client_mysql.get_balance("tb1pmuuv2qjujv9qawqcc424nhlnmux7mvsyqj7qgc6z0vwqujvx4k9s34kuxn")
    utxo = await btc_client_mysql.monitor._get_utxo_data("tb1pmuuv2qjujv9qawqcc424nhlnmux7mvsyqj7qgc6z0vwqujvx4k9s34kuxn")
    utxo_tx_id_list = [tx[0] for tx in utxo]
    assert "0cfe0b554f766ed374a426497a9163a3beb9732e799ffa623894efaea81a4415" in utxo_tx_id_list
    assert "56f677b37c44f9609f6d2ee121effe5a59b0c550c9be5073d03d8c365dde0e83" in utxo_tx_id_list
    assert "39169a8dcb800afcd71adbbca93f209d5006eb787c98668e22c0550a8450f202" in utxo_tx_id_list
    assert "5f8c9b318ab9156598d3d04701c08f374832bb1ce8704e629a68961b08d290c1" in utxo_tx_id_list
    assert "a9bfc94a0a5d7e7e46f1626b9d1848da96cc85689528ca3187e55164be8f3d04" in utxo_tx_id_list

    assert balance == 499747253
    
    balance = await btc_client_mysql.get_balance("tb1putn7zkfjr97xd77as09w5syy3x7xr5crp99wycjcxgdwrq3le9lse9fdkd")
    
    assert balance == 36481286

    balance = await btc_client_mysql.get_balance("tb1p2z70my9zwxsuqevdxjn909c7jyp89z4qv3uhy5wggn47ruvfa65s4fnngd")
    utxo = await btc_client_mysql.monitor._get_utxo_data("tb1p2z70my9zwxsuqevdxjn909c7jyp89z4qv3uhy5wggn47ruvfa65s4fnngd")
    assert balance == 478778279

    end_balance_utxo_wallet = await btc_client_mysql.get_balance(
        "tb1paf6damf5052arl3r2lsufuhyu48yth8mrgxdtqj9pervz624q6xqxm7ew3"
    )
    assert end_balance_utxo_wallet == 0
    

@vcr_c.use_cassette("btc/test_async_monitoring.yaml")
async def test_async_monitoring(btc_client: AioTxBTCClient):
    blocks = []
    transactions = []

    @btc_client.monitor.on_block
    async def handle_block(block):
        blocks.append(block)

    @btc_client.monitor.on_transaction
    async def handle_transaction(transaction):
        transactions.append(transaction)

    await btc_client.import_address("tb1pmuuv2qjujv9qawqcc424nhlnmux7mvsyqj7qgc6z0vwqujvx4k9s34kuxn", 2811502)
    await btc_client.import_address("tb1putn7zkfjr97xd77as09w5syy3x7xr5crp99wycjcxgdwrq3le9lse9fdkd")
    await btc_client.import_address("tb1p2z70my9zwxsuqevdxjn909c7jyp89z4qv3uhy5wggn47ruvfa65s4fnngd")
    await btc_client.import_address("tb1paf6damf5052arl3r2lsufuhyu48yth8mrgxdtqj9pervz624q6xqxm7ew3")
    await btc_client.monitor._add_new_utxo(
        "tb1paf6damf5052arl3r2lsufuhyu48yth8mrgxdtqj9pervz624q6xqxm7ew3",
        "8b88ebe1d0bb1eda7b1512853e75d3884b99e6634317589b1f7457ab92f1e41f",
        499824213,
        0,
    )
    start_balance_utxo_wallet = await btc_client.get_balance(
        "tb1paf6damf5052arl3r2lsufuhyu48yth8mrgxdtqj9pervz624q6xqxm7ew3"
    )
    assert start_balance_utxo_wallet == 499824213
    await btc_client.start_monitoring()

    try:
        await asyncio.sleep(3)
    except KeyboardInterrupt:
        btc_client.stop_monitoring()

    assert len(blocks) > 0
    assert len(transactions) > 0

    balance = await btc_client.get_balance("tb1pmuuv2qjujv9qawqcc424nhlnmux7mvsyqj7qgc6z0vwqujvx4k9s34kuxn")
    assert balance == 333

    balance = await btc_client.get_balance("tb1putn7zkfjr97xd77as09w5syy3x7xr5crp99wycjcxgdwrq3le9lse9fdkd")
    assert balance == 36731406

    balance = await btc_client.get_balance("tb1p2z70my9zwxsuqevdxjn909c7jyp89z4qv3uhy5wggn47ruvfa65s4fnngd")
    assert balance == 478837110

    end_balance_utxo_wallet = await btc_client.get_balance(
        "tb1paf6damf5052arl3r2lsufuhyu48yth8mrgxdtqj9pervz624q6xqxm7ew3"
    )
    assert end_balance_utxo_wallet == 0

    assert 2811502 in blocks
    assert 2811503 in blocks
    assert "e6eda82369b59c805998fc554f55163d4e4d38f5817adf8c1b6b9f917937bd77" in [tx["txid"] for tx in transactions]
    assert "56f677b37c44f9609f6d2ee121effe5a59b0c550c9be5073d03d8c365dde0e83" in [tx["txid"] for tx in transactions]



# import asyncio

# from conftest import vcr_c

# from aiotx.clients import AioTxBSCClient


# @vcr_c.use_cassette("tests/fixtures/cassettes/bsc/test_async_monitoring.yaml")
# async def test_async_monitoring(bsc_client: AioTxBSCClient):
#     blocks = []
#     transactions = []

#     @bsc_client.monitor.on_block
#     async def handle_block(block):
#         blocks.append(block)

#     @bsc_client.monitor.on_transaction
#     async def handle_transaction(transaction):
#         transactions.append(transaction)

#     await bsc_client.start_monitoring(584)
#     try:
#         await asyncio.sleep(3)
#     except KeyboardInterrupt:
#         bsc_client.stop_monitoring()

#     assert len(blocks) > 0
#     assert len(transactions) > 0

#     assert 584 in blocks
#     assert 585 in blocks
#     assert "0x2a9d5aea2c8e4759e4e664f48e0dac79e76cb024e670251ddb2788144cee6f43" in [tx["hash"] for tx in transactions]
#     assert "0xa45ccbf3c820f7bec2809461a8b7f1856246ed3b650ec2840bfbc303f118b76c" in [tx["hash"] for tx in transactions]

#     for tx in transactions:
#         assert "aiotx_decoded_input" in tx.keys()



import asyncio

from conftest import vcr_c

from aiotx.clients import AioTxBSCClient


@vcr_c.use_cassette("tests/fixtures/cassettes/eth/test_async_monitoring.yaml")
async def test_async_monitoring(eth_client: AioTxBSCClient):
    blocks = []
    transactions = []

    @eth_client.monitor.on_block
    async def handle_block(block):
        blocks.append(block)

    @eth_client.monitor.on_transaction
    async def handle_transaction(transaction):
        transactions.append(transaction)

    await eth_client.start_monitoring(2834064)
    try:
        await asyncio.sleep(3)
    except KeyboardInterrupt:
        eth_client.stop_monitoring()

    assert len(blocks) > 0
    assert len(transactions) > 0

    assert 2834064 in blocks
    assert 2834065 in blocks
    assert "0x5b6fd8fda590e887531df12dec5faf2ce8c94a3eeb56bcc1fde760fabd64e56e" in [tx["hash"] for tx in transactions]
    assert "0x10f4376f7efe2637be4d012eebad143dce58626146befa48204f1275476064d6" in [tx["hash"] for tx in transactions]

    for tx in transactions:
        assert "aiotx_decoded_input" in tx.keys()



import os
import sys

import pytest
from conftest import vcr_c

from aiotx.clients import AioTxBTCClient
from aiotx.types import FeeEstimate

TEST_BTC_WALLET_PRIVATE_KEY = os.getenv("TEST_BTC_WALLET_PRIVATE_KEY")
assert TEST_BTC_WALLET_PRIVATE_KEY is not None, "add TEST_BTC_WALLET_PRIVATE_KEY"
TEST_BTC_ADDRESS = "tb1qswslzcdulvlk62gdrg8wa0sw36f938h2pyfreh"


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("btc/send_transaction.yaml")
async def test_send_transaction(btc_client: AioTxBTCClient):
    await btc_client.monitor._add_new_address(TEST_BTC_ADDRESS)
    await btc_client.monitor._add_new_utxo(
        TEST_BTC_ADDRESS, "25d56f693d5c4d00d3f98e58c8bd66e8db930e38c8a556bd67737b66cbf31ab9", 16263, 0
    )

    amount = 1000
    fee = btc_client.to_satoshi(0.00001)
    tx_id = await btc_client.send(
        TEST_BTC_WALLET_PRIVATE_KEY, "tb1q4tf08gc2vf8dgfqtd3jen3s5tddz6uu0mnyf90", amount, fee
    )
    utxo_list = await btc_client.monitor._get_utxo_data(TEST_BTC_ADDRESS)
    utxo_list = await btc_client.monitor._get_utxo_data(TEST_BTC_ADDRESS)
    assert len(utxo_list) == 1
    assert utxo_list[0][0] == "4f9a64d28ae22134379f7da50282d7ca2ead8a1f8378b20e25defcaaa5c59228"
    assert utxo_list[0][2] == 14263
    assert tx_id == "4f9a64d28ae22134379f7da50282d7ca2ead8a1f8378b20e25defcaaa5c59228"
    await btc_client.monitor._delete_utxo("25d56f693d5c4d00d3f98e58c8bd66e8db930e38c8a556bd67737b66cbf31ab9", 0)


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("btc/send_to_all_address_types.yaml")
async def test_send_to_all_address_types(btc_client: AioTxBTCClient):
    fee = btc_client.to_satoshi(0.00006)
    await btc_client.monitor._add_new_address(TEST_BTC_ADDRESS)
    await btc_client.monitor._add_new_utxo(
        TEST_BTC_ADDRESS, "25d56f693d5c4d00d3f98e58c8bd66e8db930e38c8a556bd67737b66cbf31ab9", 16263, 0
    )
    amount = 1000
    tx_id = await btc_client.send_bulk(
        TEST_BTC_WALLET_PRIVATE_KEY,
        {
            "mtkbaiLiUH3fvGJeSzuN3kUgmJzqinLejJ": amount,  # Legacy P2PKH address
            "2NDYhjHKdPLYjD5kVZVgeo8JxDrDUcRC14Q": amount,  # P2SH address
            "tb1qq4ay9zw2mygwucz47v6hsda04a9l2yajyynh00": amount,  # Bech32 address
            "tb1pwgjc67568cu6vncgve60t6eaht0rwr4pwdymnrpezrpepg3q5t8sp5250h": amount,  # Bech32M address
        },
        fee,
    )
    utxo_list = await btc_client.monitor._get_utxo_data(TEST_BTC_ADDRESS)
    assert len(utxo_list) == 1
    assert utxo_list[0][0] == "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2"
    assert utxo_list[0][2] == 6263
    assert tx_id == "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2"


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("btc/send_with_auto_fee.yaml")
async def test_send_with_auto_fee(btc_client: AioTxBTCClient):

    await btc_client.monitor._add_new_address(TEST_BTC_ADDRESS)
    await btc_client.monitor._add_new_utxo(
        TEST_BTC_ADDRESS, " 35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ", 6263, 0
    )
    amount = 1000
    tx_id = await btc_client.send_bulk(
        TEST_BTC_WALLET_PRIVATE_KEY,
        {"mtkbaiLiUH3fvGJeSzuN3kUgmJzqinLejJ": amount},
        conf_target=1,
        estimate_mode=FeeEstimate.CONSERVATIVE,
    )
    
    assert tx_id == "77cb20c4b0325242b9e5f45f4850e5387dc585d6b72bb36ba65a126534436973"

    utxo_list = await btc_client.monitor._get_utxo_data(TEST_BTC_ADDRESS)
    assert len(utxo_list) == 0




async def test_get_balance(btc_client: AioTxBTCClient):
    await btc_client.monitor._add_new_address(TEST_BTC_ADDRESS)
    await btc_client.monitor._add_new_utxo(
        TEST_BTC_ADDRESS, "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ", 6263, 0
    )

    balance = await btc_client.get_balance(TEST_BTC_ADDRESS)
    assert balance == 6263

    await btc_client.monitor._delete_utxo("35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ", 0)

    balance = await btc_client.get_balance(TEST_BTC_ADDRESS)
    assert balance == 0

    await btc_client.monitor._add_new_utxo(
        TEST_BTC_ADDRESS, "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ", 6263, 0
    )
    balance = await btc_client.get_balance(TEST_BTC_ADDRESS)
    assert balance == 6263

    await btc_client.monitor._mark_utxo_used("35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ", 0)

    balance = await btc_client.get_balance(TEST_BTC_ADDRESS)
    assert balance == 0
        


@pytest.mark.mysql
async def test_get_balance_mysql(btc_client_mysql: AioTxBTCClient):
    await btc_client_mysql.monitor._add_new_address(TEST_BTC_ADDRESS)
    await btc_client_mysql.monitor._add_new_utxo(
        TEST_BTC_ADDRESS, "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ", 6263, 0
    )

    balance = await btc_client_mysql.get_balance(TEST_BTC_ADDRESS)
    assert balance == 6263

    await btc_client_mysql.monitor._delete_utxo("35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ", 0)

    balance = await btc_client_mysql.get_balance(TEST_BTC_ADDRESS)
    assert balance == 0

    await btc_client_mysql.monitor._add_new_utxo(
        TEST_BTC_ADDRESS, "35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ", 6263, 0
    )
    balance = await btc_client_mysql.get_balance(TEST_BTC_ADDRESS)
    assert balance == 6263

    await btc_client_mysql.monitor._mark_utxo_used("35158b7a8b6057bc67f6d904c64b5986adea8260f0bc96cbd755b530878e3cc2 ", 0)

    balance = await btc_client_mysql.get_balance(TEST_BTC_ADDRESS)
    assert balance == 0




import asyncio
import os
import sys

import pytest
from conftest import vcr_c

from aiotx.clients import AioTxLTCClient
from aiotx.exceptions import InsufficientFunds

TEST_LTC_WALLET_PRIVATE_KEY = os.getenv("TEST_LTC_WALLET_PRIVATE_KEY")
assert TEST_LTC_WALLET_PRIVATE_KEY is not None, "add TEST_LTC_WALLET_PRIVATE_KEY"
TEST_LTC_ADDRESS = "tltc1qswslzcdulvlk62gdrg8wa0sw36f938h2cvtaf7"


@vcr_c.use_cassette("ltc/get_last_block.yaml")
async def test_get_last_block(ltc_public_client: AioTxLTCClient):
    block_id = await ltc_public_client.get_last_block_number()
    assert isinstance(block_id, int)


@vcr_c.use_cassette("ltc/get_block_by_number.yaml")
async def test_get_block_by_number(ltc_public_client: AioTxLTCClient):
    block = await ltc_public_client.get_block_by_number(3247846)
    assert isinstance(block, dict)

    assert block["hash"] == "4314081d2a5d8633c51799113aac1516f174d5da5793c966848ac34177fb61c9"

    tx_hashes = [tx["hash"] for tx in block["tx"]]
    assert "68e73174f7b44edad7289a771d3069ff91c8bbb30cf96ea9861a4a312c1a2dda" in tx_hashes
    assert "4676f0e42c940f827c7f3b580119c10fe282b39bdbb798dc96d9292c387b37f0" in tx_hashes


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("ltc/send_transaction.yaml")
async def test_send_transaction(ltc_public_client: AioTxLTCClient):
    await ltc_public_client.monitor._add_new_address(TEST_LTC_ADDRESS)
    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "55863cc61de0c6c1c87282d3d6fb03650c0fc90ed3282191c618069cbde1d525", 39000000, 0
    )

    amount = ltc_public_client.to_satoshi(0.1)
    fee = ltc_public_client.to_satoshi(0.005)
    tx_id = await ltc_public_client.send(
        TEST_LTC_WALLET_PRIVATE_KEY, "tltc1q24gng65qj3wr55878324w2eeeta4k2plfwaf54", amount, fee
    )
    assert tx_id == "a006aedf3a08f423434aa781988997a0526f9365fe228fb8934ea64bbbb9d055"
    utxo_list = await ltc_public_client.monitor._get_utxo_data(TEST_LTC_ADDRESS)
    assert len(utxo_list) == 1
    assert utxo_list[0][0] == "a006aedf3a08f423434aa781988997a0526f9365fe228fb8934ea64bbbb9d055"
    assert utxo_list[0][2] == 28500000
    await ltc_public_client.monitor._delete_utxo("55863cc61de0c6c1c87282d3d6fb03650c0fc90ed3282191c618069cbde1d525", 0)


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("ltc/send_bulk.yaml")
async def test_bulk_send_transaction(ltc_public_client: AioTxLTCClient):
    fee = ltc_public_client.to_satoshi(0.005)
    await ltc_public_client.monitor._add_new_address(TEST_LTC_ADDRESS)
    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "55863cc61de0c6c1c87282d3d6fb03650c0fc90ed3282191c618069cbde1d525", 30000000, 2
    )

    amount = ltc_public_client.to_satoshi(0.05) - fee / 2
    tx_id = await ltc_public_client.send_bulk(
        TEST_LTC_WALLET_PRIVATE_KEY,
        {
            "tltc1q24gng65qj3wr55878324w2eeeta4k2plfwaf54": amount,
            "tltc1qshy0jeejm4pw3ep4cedc5vlmxyz348epnk7etf": amount,
        },
        fee,
    )
    utxo_list = await ltc_public_client.monitor._get_utxo_data(TEST_LTC_ADDRESS)
    assert len(utxo_list) == 1
    assert utxo_list[0][0] == "64a89e7e269469c126e96d1de7b553850716c71d9081b153429ff781758a59a1"
    assert utxo_list[0][2] == 20000000
    assert tx_id == "64a89e7e269469c126e96d1de7b553850716c71d9081b153429ff781758a59a1"


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("ltc/send_from_two_utxo.yaml")
async def test_send_from_two_utxo(ltc_public_client: AioTxLTCClient):
    fee = ltc_public_client.to_satoshi(0.005)
    await ltc_public_client.monitor._add_new_address(TEST_LTC_ADDRESS)
    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "a006aedf3a08f423434aa781988997a0526f9365fe228fb8934ea64bbbb9d055", 28500000, 0
    )
    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "64a89e7e269469c126e96d1de7b553850716c71d9081b153429ff781758a59a1", 20000000, 0
    )
    amount = 28500000 + 20000000 - fee
    tx_id = await ltc_public_client.send(TEST_LTC_WALLET_PRIVATE_KEY, TEST_LTC_ADDRESS, amount, fee)
    utxo_list = await ltc_public_client.monitor._get_utxo_data(TEST_LTC_ADDRESS)
    assert len(utxo_list) == 1
    assert utxo_list[0][0] == "3fcd11698664ffea5fe00adef6bd2c1c35de66c3fafb50f9076c74ff13fea139"
    assert utxo_list[0][2] == 48000000
    assert tx_id == "3fcd11698664ffea5fe00adef6bd2c1c35de66c3fafb50f9076c74ff13fea139"


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("ltc/send_to_legacy_address.yaml")
async def test_send_to_legacy_address(ltc_public_client: AioTxLTCClient):
    fee = ltc_public_client.to_satoshi(0.005)
    await ltc_public_client.monitor._add_new_address(TEST_LTC_ADDRESS)
    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "3fcd11698664ffea5fe00adef6bd2c1c35de66c3fafb50f9076c74ff13fea139", 48000000, 0
    )
    amount = ltc_public_client.to_satoshi(0.005)
    tx_id = await ltc_public_client.send(TEST_LTC_WALLET_PRIVATE_KEY, "mq2PZs9p5ZNLbu23KLKb1tdQt1mrBJM7CX", amount, fee)
    utxo_list = await ltc_public_client.monitor._get_utxo_data(TEST_LTC_ADDRESS)
    assert len(utxo_list) == 1
    assert utxo_list[0][0] == "141c30ea6326ab447423465d2a7f4c3067812f06ef1e505c0443e85c06ed684a"
    assert utxo_list[0][2] == 47000000
    assert tx_id == "141c30ea6326ab447423465d2a7f4c3067812f06ef1e505c0443e85c06ed684a"


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("ltc/send_to_legacy_and_segwit_addresses.yaml")
async def test_send_to_legacy_and_segwit_address(ltc_public_client: AioTxLTCClient):
    fee = ltc_public_client.to_satoshi(0.005)
    await ltc_public_client.monitor._add_new_address(TEST_LTC_ADDRESS)
    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "141c30ea6326ab447423465d2a7f4c3067812f06ef1e505c0443e85c06ed684a", 47000000, 0
    )
    amount = ltc_public_client.to_satoshi(0.005)
    tx_id = await ltc_public_client.send_bulk(
        TEST_LTC_WALLET_PRIVATE_KEY,
        {"mq2PZs9p5ZNLbu23KLKb1tdQt1mrBJM7CX": amount, "tltc1q24gng65qj3wr55878324w2eeeta4k2plfwaf54": amount},
        fee,
    )
    utxo_list = await ltc_public_client.monitor._get_utxo_data(TEST_LTC_ADDRESS)
    assert len(utxo_list) == 1
    assert utxo_list[0][0] == "33f67e7ac0dde523598f416f8efa5928d9e8a4a681db48f7df7d174701225dd0"
    assert utxo_list[0][2] == 45500000
    assert tx_id == "33f67e7ac0dde523598f416f8efa5928d9e8a4a681db48f7df7d174701225dd0"


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("ltc/send_with_auto_fee.yaml")
async def test_send_with_auto_fee(ltc_public_client: AioTxLTCClient):
    await ltc_public_client.monitor._add_new_address(TEST_LTC_ADDRESS)
    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "33f67e7ac0dde523598f416f8efa5928d9e8a4a681db48f7df7d174701225dd0", 45500000, 0
    )
    amount = ltc_public_client.to_satoshi(0.005)
    tx_id = await ltc_public_client.send(TEST_LTC_WALLET_PRIVATE_KEY, "mq2PZs9p5ZNLbu23KLKb1tdQt1mrBJM7CX", amount)
    utxo_list = await ltc_public_client.monitor._get_utxo_data(TEST_LTC_ADDRESS)
    assert len(utxo_list) == 1
    assert utxo_list[0][0] == "ac3c62cb37887a41235fecbc8dd22c8a8b5b74e1d2695dbc78e6d05a0cbdf2e9"
    assert utxo_list[0][2] == 44999715
    assert tx_id == "ac3c62cb37887a41235fecbc8dd22c8a8b5b74e1d2695dbc78e6d05a0cbdf2e9"


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("ltc/send_with_auto_fee_and_deduct_commission.yaml")
async def test_send_with_auto_fee_and_deduct_commission(ltc_public_client: AioTxLTCClient):
    await ltc_public_client.monitor._add_new_address(TEST_LTC_ADDRESS)
    utxo_amount = ltc_public_client.to_satoshi(0.44448584)
    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "487933e5f8028e9235f76dacb368efd64bb6f038989ff92b217661c192f0055f", utxo_amount, 0
    )
    amount = ltc_public_client.to_satoshi(0.0055)
    tx_id = await ltc_public_client.send(TEST_LTC_WALLET_PRIVATE_KEY, "mq2PZs9p5ZNLbu23KLKb1tdQt1mrBJM7CX", amount, deduct_fee=True)
    utxo_list = await ltc_public_client.monitor._get_utxo_data(TEST_LTC_ADDRESS)
    assert len(utxo_list) == 1
    assert utxo_list[0][0] == "3ed6c7a8e3b263679c47223f3e9c65721f865378e1b6799182f0607e1ebf9179"
    assert utxo_list[0][2] == 43898584
    assert tx_id == "3ed6c7a8e3b263679c47223f3e9c65721f865378e1b6799182f0607e1ebf9179"

@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("ltc/bulk_send_with_auto_fee_and_deduct_commission.yaml")
async def test_bulk_send_with_auto_fee_and_deduct_commission(ltc_public_client: AioTxLTCClient):
    await ltc_public_client.monitor._add_new_address(TEST_LTC_ADDRESS)
    utxo_amount = ltc_public_client.to_satoshi(0.43397453)
    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "3f7e5e1a897b6698d551f2b576d15eda85bf1c0d2706f7cb8c4b17a75632d1d2", utxo_amount, 0
    )
    amount = ltc_public_client.to_satoshi(0.005)
    tx_id = await ltc_public_client.send_bulk(TEST_LTC_WALLET_PRIVATE_KEY, {"mq2PZs9p5ZNLbu23KLKb1tdQt1mrBJM7CX": amount,
                                                                            "tltc1qshy0jeejm4pw3ep4cedc5vlmxyz348epnk7etf": amount},
                                                                            deduct_fee=True)
    
    utxo_list = await ltc_public_client.monitor._get_utxo_data(TEST_LTC_ADDRESS)
    assert len(utxo_list) == 1
    assert utxo_list[0][0] == "e328d42e61f6b022c1534cd2e7e184180574172c988bea359f0b8e8734f178c6"
    assert utxo_list[0][2] == 42397453
    assert tx_id == "e328d42e61f6b022c1534cd2e7e184180574172c988bea359f0b8e8734f178c6"


async def test_zero_balance_error(ltc_public_client: AioTxLTCClient):
    await ltc_public_client.monitor._add_new_address(TEST_LTC_ADDRESS)
    amount = ltc_public_client.to_satoshi(0.005)
    with pytest.raises(InsufficientFunds) as excinfo:  
        await ltc_public_client.send(TEST_LTC_WALLET_PRIVATE_KEY, "mq2PZs9p5ZNLbu23KLKb1tdQt1mrBJM7CX", amount)
    assert str(excinfo.value) == "We have only 0 satoshi and it's 500000 at least needed to cover that transaction!"
    

async def test_not_enough_balance_error(ltc_public_client: AioTxLTCClient):
    await ltc_public_client.monitor._add_new_address(TEST_LTC_ADDRESS)
    utxo_amount = ltc_public_client.to_satoshi(0.005)
    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "3f7e5e1a897b6698d551f2b576d15eda85bf1c0d2706f7cb8c4b17a75632d1d2", utxo_amount, 0
    )
    amount = ltc_public_client.to_satoshi(0.01)
    with pytest.raises(InsufficientFunds) as excinfo:  
        await ltc_public_client.send(TEST_LTC_WALLET_PRIVATE_KEY, "mq2PZs9p5ZNLbu23KLKb1tdQt1mrBJM7CX", amount)
    assert str(excinfo.value) == "We have only 500000 satoshi and it's 1000000 at least needed to cover that transaction!"



@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("ltc/send_few_single_transactions.yaml")
async def test_send_few_single_transactions(ltc_public_client: AioTxLTCClient):
    await ltc_public_client.monitor._add_new_address(TEST_LTC_ADDRESS)
    utxo_amount = ltc_public_client.to_satoshi(0.005)
    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "89596bf8624118c6c3cd8cc2ebb6a2e19bc1c97abf994de9f23b8a2aef6765c1", utxo_amount, 2
    )
    amount = ltc_public_client.to_satoshi(0.01)
    with pytest.raises(InsufficientFunds) as excinfo:  
        await ltc_public_client.send(TEST_LTC_WALLET_PRIVATE_KEY, TEST_LTC_ADDRESS, amount)
    assert str(excinfo.value) == "We have only 500000 satoshi and it's 1000000 at least needed to cover that transaction!"

    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "c2f8e76c9537e5defa921b31edeb269603b8adec025c04d8e8b0e5764b80b861", utxo_amount, 2
    )
    
    tx_id = await ltc_public_client.send(TEST_LTC_WALLET_PRIVATE_KEY, TEST_LTC_ADDRESS, amount, deduct_fee=True)
    utxo_list = await ltc_public_client.monitor._get_utxo_data(TEST_LTC_ADDRESS)
    assert len(utxo_list) == 1
    assert utxo_list[0][0] == "477ff66cdf3c97661c66d38a05cfae018d0358fec300a8b85bde65808e6cf8f9"
    assert utxo_list[0][2] == 997636
    assert tx_id == "477ff66cdf3c97661c66d38a05cfae018d0358fec300a8b85bde65808e6cf8f9"

    await ltc_public_client.monitor._mark_utxo_used("477ff66cdf3c97661c66d38a05cfae018d0358fec300a8b85bde65808e6cf8f9", 0)

    with pytest.raises(InsufficientFunds) as excinfo:  
        await ltc_public_client.send(TEST_LTC_WALLET_PRIVATE_KEY, TEST_LTC_ADDRESS, amount)
    assert str(excinfo.value) == "We have only 0 satoshi and it's 1000000 at least needed to cover that transaction!"

    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "8117d7261873ab0bb49e195043d70693362cebe3275333678da5f738932bb3a7", utxo_amount, 2
    )

    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "ad83a8dedeb35659cfc65053021125618ebdd0f60ee3b946fda5ea77df6a0047", utxo_amount, 2
    )

    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "a51a47a487ab537dbf263d70a4ffd4535b57f479ba792c1549c4a4e9a44fcfb2", utxo_amount, 2
    )
    second_tx_amount = ltc_public_client.to_satoshi(0.012)
    tx_id = await ltc_public_client.send(TEST_LTC_WALLET_PRIVATE_KEY, TEST_LTC_ADDRESS, second_tx_amount)
    utxo_list = await ltc_public_client.monitor._get_utxo_data(TEST_LTC_ADDRESS)
    assert len(utxo_list) == 2
    balance = await ltc_public_client.get_balance(TEST_LTC_ADDRESS)
    assert balance == 296321 + 1200000
    assert tx_id == "09dac3c34ce8013a074890cdc34f90e264fda192b995350b7ad6d283f7d9276d"


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("ltc/send_with_fee_per_byte.yaml")
async def test_send_with_fee_per_byte(ltc_public_client: AioTxLTCClient):
    await ltc_public_client.monitor._add_new_address(TEST_LTC_ADDRESS)
    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "45d11c95718b39f52e11cddd69ce5be0585398afa299f0d7ab1e1ef7aa48ce5d", 50000, 1
    )
    tx_id = await ltc_public_client.send(TEST_LTC_WALLET_PRIVATE_KEY, "mq2PZs9p5ZNLbu23KLKb1tdQt1mrBJM7CX", 50000, deduct_fee=True, fee_per_byte=201)
    utxo_list = await ltc_public_client.monitor._get_utxo_data(TEST_LTC_ADDRESS)
    assert len(utxo_list) == 0
    assert tx_id == "d2ed7d677005319439b4eb1a17cd66bd4cf32eea98d2b951352d627e4be9e340"


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@vcr_c.use_cassette("ltc/send_with_fee_per_byte2.yaml")
async def test_send_with_fee_per_byte2(ltc_public_client: AioTxLTCClient):
    await ltc_public_client.monitor._add_new_address(TEST_LTC_ADDRESS)
    await ltc_public_client.monitor._add_new_utxo(
        TEST_LTC_ADDRESS, "477ff66cdf3c97661c66d38a05cfae018d0358fec300a8b85bde65808e6cf8f9", 997636, 0
    )
    tx_id = await ltc_public_client.send(TEST_LTC_WALLET_PRIVATE_KEY, "mq2PZs9p5ZNLbu23KLKb1tdQt1mrBJM7CX", 997636, deduct_fee=True, fee_per_byte=505)
    utxo_list = await ltc_public_client.monitor._get_utxo_data(TEST_LTC_ADDRESS)
    assert len(utxo_list) == 0
    assert tx_id == "b9d298d1af00740332d663d2902d30453781447e13ef0cc6aa07d2922df12f61"



@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Skipping transaction signing tests on Windows because we are not using RFC6979 from fastecdsa by default",
)
@pytest.mark.mysql
@vcr_c.use_cassette("ltc/monitoring_balance_send_mark_as_used.yaml")
async def test_monitoring_balance_send_mark_as_used(ltc_client_mysql: AioTxLTCClient):
    """
    Mega test :D It's for getting UTXO from monitoring (how it will be done by user)
    Checking balance, getting and checking UTXO list and sending transaction, checking what UTXO's was flagged
    as used
    """

    blocks = []
    transactions = []

    @ltc_client_mysql.monitor.on_block
    async def handle_block(block):
        blocks.append(block)

    @ltc_client_mysql.monitor.on_transaction
    async def handle_transaction(transaction):
        transactions.append(transaction)
    
    await ltc_client_mysql.import_address(TEST_LTC_ADDRESS, 3262692)
    await ltc_client_mysql.start_monitoring()

    try:
        await asyncio.sleep(2)
    except KeyboardInterrupt:
        ltc_client_mysql.stop_monitoring()

    assert len(blocks) > 0
    assert len(transactions) > 0
    assert 3262692 in blocks

    assert "a51a47a487ab537dbf263d70a4ffd4535b57f479ba792c1549c4a4e9a44fcfb2" in [tx["txid"] for tx in transactions]

    # Removing some transaction from block 3262692 because we know they was used, but not monitoring in that test case
    await ltc_client_mysql.monitor._mark_utxo_used("8117d7261873ab0bb49e195043d70693362cebe3275333678da5f738932bb3a7", 2)
    await ltc_client_mysql.monitor._mark_utxo_used("89596bf8624118c6c3cd8cc2ebb6a2e19bc1c97abf994de9f23b8a2aef6765c1", 2)
    await ltc_client_mysql.monitor._mark_utxo_used("a51a47a487ab537dbf263d70a4ffd4535b57f479ba792c1549c4a4e9a44fcfb2", 2)
    await ltc_client_mysql.monitor._mark_utxo_used("ad83a8dedeb35659cfc65053021125618ebdd0f60ee3b946fda5ea77df6a0047", 2)
    await ltc_client_mysql.monitor._mark_utxo_used("c2f8e76c9537e5defa921b31edeb269603b8adec025c04d8e8b0e5764b80b861", 2)

    tx_id = await ltc_client_mysql.send(TEST_LTC_WALLET_PRIVATE_KEY, "mq2PZs9p5ZNLbu23KLKb1tdQt1mrBJM7CX", 10000, deduct_fee=True)
    assert tx_id == "e9830d6d57c3a77a8d64c0df1a44eccdb0eeed6ef8fda66fb82d113c45dff6a1"

    


import asyncio

from conftest import vcr_c

from aiotx.clients import AioTxLTCClient


@vcr_c.use_cassette("tests/fixtures/cassettes/ltc/test_async_monitoring.yaml")
async def test_async_monitoring(ltc_public_client: AioTxLTCClient):
    blocks = []
    transactions = []

    @ltc_public_client.monitor.on_block
    async def handle_block(block):
        blocks.append(block)

    @ltc_public_client.monitor.on_transaction
    async def handle_transaction(transaction):
        transactions.append(transaction)

    await ltc_public_client.import_address("tltc1qsawz44ppfnxmnat7635f83exgf9mynrzs5tsgl", 3247853)
    await ltc_public_client.start_monitoring()

    try:
        await asyncio.sleep(3)
    except KeyboardInterrupt:
        ltc_public_client.stop_monitoring()

    assert len(blocks) > 0
    assert len(transactions) > 0

    assert 3247853 in blocks
    assert 3247854 in blocks
    assert "7604930759225ab74868969da6b19b399d8b189bd9506f39e15019f8e08a1d44" in [tx["txid"] for tx in transactions]
    assert "992b887a10aa924430e905174535fd92733136b76d39e835c109e3ff2719bd69" in [tx["txid"] for tx in transactions]



import json

import pkg_resources

from ._evm_base_client import AioTxEVMClient
from ._utxo_base_client import AioTxUTXOClient


class AioTxBSCClient(AioTxEVMClient):
    def __init__(self, node_url):
        super().__init__(node_url)
        bep20_abi_json = pkg_resources.resource_string('aiotx.utils', 'bep20_abi.json')
        self._bep20_abi = json.loads(bep20_abi_json)

    def _get_abi_entries(self):
        return [entry for entry in self._bep20_abi]


class AioTxETHClient(AioTxEVMClient):
    def __init__(self, node_url):
        super().__init__(node_url)
        erc20_abi_json = pkg_resources.resource_string('aiotx.utils', 'erc20_abi.json')
        self._erc20_abi = json.loads(erc20_abi_json)

    def _get_abi_entries(self):
        return [entry for entry in self._erc20_abi]


class AioTxBTCClient(AioTxUTXOClient):

    def __init__(
        self,
        node_url,
        testnet=False,
        node_username: str = "",
        node_password: str = "",
        db_url="sqlite+aiosqlite:///aiotx_utxo.sqlite",
    ):
        network_name = "testnet" if testnet else "bitcoin"
        super().__init__(node_url, testnet, node_username, node_password, network_name, db_url)


class AioTxLTCClient(AioTxUTXOClient):

    def __init__(
        self,
        node_url,
        testnet=False,
        node_username: str = "",
        node_password: str = "",
        db_url="sqlite+aiosqlite:///aiotx_utxo.sqlite",
    ):
        network_name = "litecoin_testnet" if testnet else "litecoin"
        super().__init__(node_url, testnet, node_username, node_password, network_name, db_url)
